/*
Write talk back command and fetch command from talk back

Copyright 2023, LN.

*/

#include <WiFiS3.h>

#define LED_BUILTIN 13
#define SWITCH 5

// Update these with values suitable for your network.
#include "arduino_secrets.h" 
///////please enter your sensitive data in the Secret tab/arduino_secrets.h
char ssid[] = SECRET_SSID;        // your network SSID (name)
char pass[] = SECRET_PASS;    // your network password (use for WPA, or use as key for WEP)
int status = WL_IDLE_STATUS;     // the WiFi radio's status


const char *API = CHANNEL_API_KEY;  // CHANGE ME
const char *HOST = HOST_ADD;
const int PORT = PORT_NUMBER;
const char *field1 = "field1";

WiFiClient client;

unsigned long myTalkBackID = TALK_BACKID;
const char *myTalkBackKey = TALK_BACK_AKEY;

bool tSwitch, pSwitch;
bool tbNumber=true;
bool pn=true;

void setup_wifi() {
  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);

  // check for the WiFi module:
  if (WiFi.status() == WL_NO_MODULE) {
    Serial.println("Communication with WiFi module failed!");
    // don't continue
    while (true);
  }

  String fv = WiFi.firmwareVersion();
  if (fv < WIFI_FIRMWARE_LATEST_VERSION) {
    Serial.println("Please upgrade the firmware");
  }

  // attempt to connect to WiFi network:
  while (status != WL_CONNECTED) {
    Serial.print("Attempting to connect to WPA SSID: ");
    Serial.println(ssid);
    // Connect to WPA/WPA2 network:
    if(strlen(pass)) status = WiFi.begin(ssid, pass);
    else status = WiFi.begin(ssid);

    // wait 1 second for connection:
    delay(1000);
  }

  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());
}

//In the setup function, initialize the LED, start the serial monitor and 
//establishing a connection to the local WiFi network.
void setup() {
  pinMode(LED_BUILTIN, OUTPUT);  // Set up LED
  pinMode(SWITCH, INPUT_PULLUP);  // Set up Switch  

  Serial.begin(115200);  // Initialize serial
  setup_wifi();

  Serial.println("");
  Serial.println("\nConnected.");
}


//In the main loop, start by createint the POST message with the correct parameters. Send the POST request, verify the result,
//and check for a TalkBack command. Wait 60 seconds and check the queue again.
void loop() {
  //POST request strings
  String postMessage;
  String tbURI;
  String newCommand = String();
  int x=-1000;

  tbURI = String("/talkbacks/") + String(myTalkBackID);
  postMessage = String("api_key=") + String(myTalkBackKey);


  //Switch reading
  tSwitch = digitalRead(SWITCH);

  if (!tSwitch && pSwitch) {
      Serial.print(tSwitch); Serial.print(" "); Serial.println(pSwitch);
    // Create the TalkBack URI
    tbURI += String("/commands");

    // Create the message body for the POST out of the values
    if(pn)postMessage += String("&command_string=LED_ON");
    else postMessage += String("&command_string=LED_OFF");
    pn=!pn;
    Serial.println("Poslah LED_ON!");

    // Make the POST to ThingSpeak
    x = httpPOST(tbURI, postMessage, newCommand);

    tbNumber=true;
  } 
  else if(tbNumber){
    Serial.print(tSwitch); Serial.print(" "); Serial.println(pSwitch);
    // Create the TalkBack URI
    tbURI += String("/commands/execute");

    // Make the POST to ThingSpeak
    x = httpPOST(tbURI, postMessage, newCommand);

    tbNumber=false;
  }

  // Check the result
  if (x == 200) {
    Serial.println("checking queue...");
    // check for a command returned from TalkBack
    if (newCommand.length() != 0) {
      Serial.print("  Latest command from queue: ");
      Serial.println(newCommand);
      if (newCommand.indexOf("LED_ON") != (-1)) {
        digitalWrite(LED_BUILTIN, HIGH);
        tbNumber=true;
      }

      if (newCommand.indexOf("LED_OFF") != -1) {
        digitalWrite(LED_BUILTIN, LOW);
        tbNumber=true;
      }
    } else {
      Serial.println("  Nothing new.");
    }

  } else if(x!=-1000){
    Serial.println("Problem checking queue. HTTP error code " + String(x));
  }

  pSwitch=tSwitch;
}



//Get sensor value
int getSensorData() {
  return random(1000);  // Replace with
}


// Globalne promenljive za backoff logiku
static bool backoffActive = false;
static unsigned long backoffStartTime = 0;

int httpPOST(String uri, String postMessage, String &response) {
  unsigned long initialBackoff = 16000; // 16 sekundi početno
  unsigned long backoffDuration = initialBackoff;
  const float backoffFactor = 1.5; // Svaki put povećaj za 50%
  const int maxRetries = 5; // Maksimalno 5 pokušaja
  int retryCount = 0;

  bool connectSuccess = false;

  do {
    Serial.println("Povezivanje sa serverom...");
    connectSuccess = client.connect("api.thingspeak.com", 80);
    
    if (!connectSuccess) {
      Serial.print("Neuspešno povezivanje! Čekam ");
      Serial.print(backoffDuration / 1000);
      Serial.println(" sekundi pre novog pokušaja...");
      
      unsigned long waitStart = millis();
      while (millis() - waitStart < backoffDuration) {
        delay(100);
      }

      backoffDuration = (unsigned long)(backoffDuration * backoffFactor); // Povećaj backoff
      retryCount++;
      if (retryCount >= maxRetries) {
        Serial.println("Prekoračen maksimalni broj pokušaja povezivanja!");
        return -301; // Odustajemo nakon maxRetries neuspeha
      }
    }
  } while (!connectSuccess);

  Serial.println("Uspešno povezivanje!");

  // Resetuj backoff trajanje nakon uspešnog povezivanja
  backoffDuration = initialBackoff;

  postMessage += "&headers=false";

  String Headers = String("POST ") + uri + " HTTP/1.1\r\n" +
                   "Host: api.thingspeak.com\r\n" +
                   "Content-Type: application/x-www-form-urlencoded\r\n" +
                   "Connection: close\r\n" +
                   "Content-Length: " + String(postMessage.length()) + "\r\n\r\n";

  Serial.println("Slanje HEADERS:");
  Serial.println(Headers);
  client.print(Headers);
  Serial.println("Završeno slanje HEADERS.");

  Serial.println("Slanje POST poruke:");
  Serial.println(postMessage);
  client.print(postMessage);
  Serial.println("Završeno slanje POST poruke.");

  // Timeout za čitanje odgovora
  client.setTimeout(5000);

  Serial.println("Čekanje na dostupne podatke...");
  long startWaitForResponseAt = millis();
  while (client.available() == 0 && millis() - startWaitForResponseAt < 5000) {
    delay(100);
  }
  Serial.println("Završena while petlja.");

  if (client.available() == 0) {
    Serial.println("Nema odgovora od servera u roku.");
    return -304;
  }

  Serial.println("Traženje HTTP/1.1 odgovora...");
  if (!client.find("HTTP/1.1 ")) { 
    Serial.println("HTTP/1.1 nije pronađen.");
    return -303;
  }
  Serial.println("Pronađen HTTP/1.1 odgovor!");

  int status = client.readStringUntil(' ').toInt();
  Serial.print("Status odgovora: ");
  Serial.println(status);

  if (status != 200) {
    Serial.println("Server vratio grešku!");
    return status;
  }

  Serial.println("Traženje kraja HTTP zaglavlja...");
  if (!client.find("\n\r\n")) {
    Serial.println("Kraj HTTP zaglavlja nije pronađen.");
    return -303;
  }
  Serial.println("Pronađen kraj HTTP zaglavlja.");

  // 🛠️ Ručno čitanje tela odgovora
  Serial.println("Čitanje tela odgovora (ručno)...");
  response = "";
  unsigned long readStart = millis();
  while (client.connected() && (millis() - readStart < 5000)) {
    while (client.available()) {
      char c = client.read();
      response += c;
      readStart = millis();
    }
  }
  Serial.println("Završeno čitanje tela odgovora.");

  return status;
}



